;;; This file eratosthenese.el was coded by Robert Boyer,
;;; roberstephenboyer@gmail.com, in 2024, and is public domain.

;;;                           A Sieve of Eratosthenese

;;; Born c. 276 B. C. in Libya, Eratosthenese calculated the radius of the Earth
;;; and was the librarian of Alexandria.

;;; The build a sieve of Eratosthenese of length n, call (build-sieve n),
;;; after which i is prime if and only if (is-a-prime i).

;;; (build-sieve (expt 10 8)) takes 1600 seconds to execute.

(defvar test-form '(progn (emacs-lisp-native-compile-and-load) (benchmark (build-sieve (expt 10 8)) 1)))

(cl-declaim (optimize (safety 0) (speed 3) (debug 0) (space 0)))

(defvar *sieve-length* 0)

(setq *sieve-length* 0)

(defvar *sieve* (make-vector 1 0))

(setq *sieve* (make-vector 1 0))

(cl-defun build-sieve (n)
  "After executing (build-sieve n), if j is a nonnegative integer less than
   *sieve-length*, then j is prime <-> (is-a-prime j)."
  (setq *sieve* (make-vector n 0))
  (setq *sieve-length* n)
  ;; Neither 0 nor 1 is prime.
  (setf (aref *sieve* 0) 1)
  (setf (aref *sieve* 1) 1)
  ;; Loop through a from 0 below l. If you encounter a 0 at file position
  ;; i, then '1 out' all the subsequent multiples of i.
  (cl-loop for i from 2 below *sieve-length* do
           (cond ((eql 0 (aref *sieve* i))
                  (cl-loop for j from (ash i 1) below *sieve-length* by i
                           do (setf (aref *sieve* j) 1))))))

(cl-defun test-sieve (n)
  (cond ((not (< n *sieve-length*)) (error "n is too big.")))
  (cl-loop for i from 2 to n do
           (let ((number-of-divisors (cl-loop for j from 2 to i when (eql 0 (% i j)) count 1)))
             (cond ((eql (aref *sieve* i) 1)
                    (cond ((not (< 1 number-of-divisors))
                           (error "bad at %s." i))))
                   ((eql (aref *sieve* i) 0)
                    (cond ((not (eql 1 number-of-divisors))
                           (error "bad at %s." i))))
                   (t (error "ridiculous")))))
  'ok)

(defun is-a-prime (n)
  (cond ((eql 0 (aref *sieve* 0))
         (error "It looks like (build-sieve) has not been run.")))
  (cond ((< n *sieve-length*)
         (eql 0 (aref *sieve* n)))
        (t (error "need a bigger sieve"))))


;;; Local Variables: ;;;
;;; mode: Emacs-Lisp ;;;
;;; coding: utf-8    ;;;
;;; End:             ;;;
